# stdlib
from typing import Any
from typing import Dict
from typing import Union

# third party
import sympy as sp

# relative
from ..common.serde.serializable import serializable
from .entity import Entity
from .scalar.gamma_scalar import GammaScalar


@serializable(recursive_serde=True)
class PrimeFactory:

    """IMPORTANT: it's very important that two tensors be able to tell that
    they are indeed referencing the EXACT same PrimeFactory. At present this is done
    by ensuring that it is literally the same python object. In the future, we will probaby
    need to formalize this. However, the main way this could go wrong is if we created some
    alternate way for checking to see if two prime factories 'sortof looked the same' but which
    in fact weren't the EXACT same object. This could lead to security leaks wherein two tensors
    think two different symbols in fact are the same symbol."""

    __attr_allowlist__ = ["prev_prime"]

    def __init__(self) -> None:
        self.prev_prime = 1

    def next(self) -> int:
        self.prev_prime = sp.nextprime(self.prev_prime)
        return self.prev_prime


@serializable(recursive_serde=True)
class VirtualMachinePrivateScalarManager:

    __attr_allowlist__ = ["prime_factory", "prime2symbol"]

    def __init__(self) -> None:
        self.prime_factory = PrimeFactory()
        self.prime2symbol: Dict[Any, Any] = {}

    def get_symbol(
        self,
        min_val: Union[bool, int, float],
        value: Union[bool, int, float],
        max_val: Union[bool, int, float],
        entity: Entity,
    ) -> int:
        gs = GammaScalar(
            min_val=min_val,
            value=value,
            max_val=max_val,
            entity=entity,
            prime=self.prime_factory.next(),
        )
        self.prime2symbol[gs.prime] = gs
        return gs.prime



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# relative
from ...... import serialize
from ......proto.core.node.common.service.resolve_pointer_type_service_pb2 import (
    ResolvePointerTypeAnswerMessage as ResolvePointerTypeAnswerMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address
from ....abstract.node import AbstractNode
from ..node_service import ImmediateNodeServiceWithReply
from .resolve_pointer_type_messages import ResolvePointerTypeMessage


@serializable()
class ResolvePointerTypeAnswerMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        type_path: str,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.type_path = type_path

    def _object2proto(self) -> ResolvePointerTypeAnswerMessage_PB:
        return ResolvePointerTypeAnswerMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            type_path=self.type_path,
        )

    @staticmethod
    def _proto2object(
        proto: ResolvePointerTypeAnswerMessage_PB,
    ) -> "ResolvePointerTypeAnswerMessage":
        return ResolvePointerTypeAnswerMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            type_path=proto.type_path,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return ResolvePointerTypeAnswerMessage_PB


class ResolvePointerTypeService(ImmediateNodeServiceWithReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: ResolvePointerTypeMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> ResolvePointerTypeAnswerMessage:
        object = node.store[msg.id_at_location]
        type_qualname = object.object_qualname
        return ResolvePointerTypeAnswerMessage(
            address=msg.reply_to, type_path=type_qualname
        )

    @staticmethod
    def message_handler_types() -> List[Type[ResolvePointerTypeMessage]]:
        return [ResolvePointerTypeMessage]



# stdlib
from typing import Any
from typing import List
from typing import Optional

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

# relative
from ..exceptions import GroupNotFoundError
from ..node_table.groups import Group
from ..node_table.usergroup import UserGroup
from .database_manager import DatabaseManager


class GroupManager(DatabaseManager):

    schema = Group
    user_group_association_schema = UserGroup

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=GroupManager.schema, db=database)
        self._association_schema = GroupManager.user_group_association_schema

    def first(self, **kwargs: Any) -> Group:
        result = super().first(**kwargs)
        if not result:
            raise GroupNotFoundError
        return result

    def create(self, group_name: str, users: Optional[List] = None) -> None:
        group_obj = self.register(name=group_name)
        if users:
            self._attach_users(group_id=group_obj.id, users=users)

    def update(
        self,
        group_id: int,
        group_name: Optional[str] = None,
        users: Optional[List] = None,
    ) -> None:
        if group_name:
            self.modify(query={"id": group_id}, values={"name": group_name})

        if users:
            self.delete_association(group=group_id)
            self._attach_users(group_id=group_id, users=users)

    def get_users(self, group_id: int) -> List[int]:
        _associations = self.db.session.query(self._association_schema).filter_by(
            group=group_id
        )
        return [assoc.user for assoc in _associations]

    def get_groups(self, user_id: int) -> List[int]:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        _associations = session_local.query(self._association_schema).filter_by(
            user=user_id
        )
        return [assoc.group for assoc in _associations]

    def contain_association(self, **kwargs: Any) -> bool:
        result = (
            self.db.session.query(self._association_schema).filter_by(**kwargs).first()
        )
        return result is not None

    def update_user_association(self, user_id: int, groups: List[int]) -> None:
        # Delete all previous group associations with this user_id
        self.delete_association(user=user_id)
        # Create new ones
        for group_id in groups:
            # Check if group exists
            if self.contain(id=group_id):
                _user_group_association = self._association_schema(
                    user=user_id, group=group_id
                )
                self.db.session.add(_user_group_association)

        self.db.session.commit()

    def _attach_users(self, group_id: int, users: List[int]) -> None:
        for user in users:
            _user_group_association = self._association_schema(
                user=user, group=group_id
            )
            self.db.session.add(_user_group_association)

        self.db.session.commit()

    def delete_association(self, **kwargs: Any) -> None:
        objects_to_delete = (
            self.db.session.query(self._association_schema).filter_by(**kwargs).all()
        )
        for obj in objects_to_delete:
            self.db.session.delete(obj)
        self.db.session.commit()



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Optional

# third party
from nacl.signing import VerifyKey
import requests
from typing_extensions import final

# relative
from .....common.serde.serializable import serializable
from ....abstract.node import AbstractNode
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage


@serializable(recursive_serde=True)
@final
class PingMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class PingReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class PingMessageWithReply(GenericPayloadMessageWithReply):
    message_type = PingMessage
    message_reply_type = PingReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            host_or_ip = str(self.kwargs["host_or_ip"])
            if not host_or_ip.startswith("http"):
                host_or_ip = f"http://{host_or_ip}"
            res = requests.get(f"{host_or_ip}/status")
            return {"host_or_ip": host_or_ip, "status_code": res.status_code}
        except Exception:
            print("Failed to run ping", self.kwargs)
            return {"host_or_ip": host_or_ip, "error": "Error"}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/run_function_or_constructor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n?proto/core/node/common/action/run_function_or_constructor.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xd6\x02\n\x1eRunFunctionOrConstructorAction\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x03(\x0c\x12X\n\x06kwargs\x18\x03 \x03(\x0b\x32H.syft.core.node.common.action.RunFunctionOrConstructorAction.KwargsEntry\x12-\n\x0eid_at_location\x18\x04 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x12%\n\x06msg_id\x18\x06 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x11\n\tis_static\x18\x07 \x01(\x08\x1a-\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x62\x06proto3'
)


_RUNFUNCTIONORCONSTRUCTORACTION = DESCRIPTOR.message_types_by_name[
    "RunFunctionOrConstructorAction"
]
_RUNFUNCTIONORCONSTRUCTORACTION_KWARGSENTRY = (
    _RUNFUNCTIONORCONSTRUCTORACTION.nested_types_by_name["KwargsEntry"]
)
RunFunctionOrConstructorAction = _reflection.GeneratedProtocolMessageType(
    "RunFunctionOrConstructorAction",
    (_message.Message,),
    {
        "KwargsEntry": _reflection.GeneratedProtocolMessageType(
            "KwargsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _RUNFUNCTIONORCONSTRUCTORACTION_KWARGSENTRY,
                "__module__": "proto.core.node.common.action.run_function_or_constructor_pb2"
                # @@protoc_insertion_point(class_scope:syft.core.node.common.action.RunFunctionOrConstructorAction.KwargsEntry)
            },
        ),
        "DESCRIPTOR": _RUNFUNCTIONORCONSTRUCTORACTION,
        "__module__": "proto.core.node.common.action.run_function_or_constructor_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.RunFunctionOrConstructorAction)
    },
)
_sym_db.RegisterMessage(RunFunctionOrConstructorAction)
_sym_db.RegisterMessage(RunFunctionOrConstructorAction.KwargsEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _RUNFUNCTIONORCONSTRUCTORACTION_KWARGSENTRY._options = None
    _RUNFUNCTIONORCONSTRUCTORACTION_KWARGSENTRY._serialized_options = b"8\001"
    _RUNFUNCTIONORCONSTRUCTORACTION._serialized_start = 166
    _RUNFUNCTIONORCONSTRUCTORACTION._serialized_end = 508
    _RUNFUNCTIONORCONSTRUCTORACTION_KWARGSENTRY._serialized_start = 463
    _RUNFUNCTIONORCONSTRUCTORACTION_KWARGSENTRY._serialized_end = 508
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.signing import VerifyKey

# relative
from .....common.message import ImmediateSyftMessageWithReply
from ....domain.domain_interface import DomainInterface
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...exceptions import RequestError
from ...exceptions import RoleNotFoundError
from ...node_table.utils import model_to_json
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..success_resp_message import SuccessResponseMessage
from .role_manager_messages import CreateRoleMessage
from .role_manager_messages import DeleteRoleMessage
from .role_manager_messages import GetRoleMessage
from .role_manager_messages import GetRoleResponse
from .role_manager_messages import GetRolesMessage
from .role_manager_messages import GetRolesResponse
from .role_manager_messages import UpdateRoleMessage

INPUT_TYPE = Union[
    Type[CreateRoleMessage],
    Type[UpdateRoleMessage],
    Type[GetRoleMessage],
    Type[GetRolesMessage],
    Type[DeleteRoleMessage],
]

INPUT_MESSAGES = Union[
    CreateRoleMessage,
    UpdateRoleMessage,
    GetRoleMessage,
    GetRolesMessage,
    DeleteRoleMessage,
]

OUTPUT_MESSAGES = Union[SuccessResponseMessage, GetRoleResponse, GetRolesResponse]


def create_role_msg(
    msg: CreateRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check key permissions
    _allowed = node.users.can_edit_roles(verify_key=verify_key)

    if not msg.name:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (name)!"
        )

    # Check if this role name was already registered
    try:
        node.roles.first(name=msg.name)
        raise RequestError(message="The role name already exists!")
    except RoleNotFoundError:
        pass

    if _allowed:
        node.roles.register(
            name=msg.name,
            can_make_data_requests=msg.can_make_data_requests,
            can_triage_data_requests=msg.can_triage_data_requests,
            can_manage_privacy_budget=msg.can_manage_privacy_budget,
            can_create_users=msg.can_create_users,
            can_manage_users=msg.can_manage_users,
            can_edit_roles=msg.can_edit_roles,
            can_manage_infrastructure=msg.can_manage_infrastructure,
            can_upload_data=msg.can_upload_data,
            can_upload_legal_document=msg.can_upload_legal_document,
            can_edit_domain_settings=msg.can_edit_domain_settings,
        )
    else:
        raise AuthorizationError("You're not allowed to create a new Role!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Role created successfully!",
    )


def update_role_msg(
    msg: UpdateRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:

    params = {
        "name": msg.name,
        "can_make_data_requests": msg.can_make_data_requests,
        "can_triage_data_requests": msg.can_triage_data_requests,
        "can_manage_privacy_budget": msg.can_manage_privacy_budget,
        "can_create_users": msg.can_create_users,
        "can_manage_users": msg.can_manage_users,
        "can_edit_roles": msg.can_edit_roles,
        "can_manage_infrastructure": msg.can_manage_infrastructure,
        "can_upload_data": msg.can_upload_data,
        "can_upload_legal_document": msg.can_upload_legal_document,
        "can_edit_domain_settings": msg.can_edit_domain_settings,
    }

    if not msg.role_id:
        raise MissingRequestKeyError

    # Check Key permissions
    _allowed = node.users.can_edit_roles(verify_key=verify_key)

    if _allowed:
        node.roles.set(role_id=msg.role_id, params=params)
    else:
        raise AuthorizationError("You're not authorized to edit this role!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Role updated successfully!",
    )


def get_role_msg(
    msg: GetRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetRoleResponse:

    # Check Key permissions
    _allowed = node.users.can_triage_requests(verify_key=verify_key)

    if _allowed:
        role = node.roles.first(id=msg.role_id)
        _msg = model_to_json(role)
    else:
        raise AuthorizationError(
            "get_role_msg You're not allowed to get User information!"
        )

    return GetRoleResponse(address=msg.reply_to, content=_msg)


def get_all_roles_msg(
    msg: GetRolesMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetRolesResponse:

    _allowed = node.users.can_triage_requests(verify_key=verify_key)

    if _allowed:
        roles = node.roles.all()
        _msg = [model_to_json(role) for role in roles]
    else:
        raise AuthorizationError("You're not allowed to get Role information!")

    return GetRolesResponse(address=msg.reply_to, content=_msg)


def del_role_msg(
    msg: DeleteRoleMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    _allowed = node.users.can_edit_roles(verify_key=verify_key)

    if _allowed:
        node.roles.delete(id=msg.role_id)
    else:
        raise AuthorizationError("You're not authorized to delete this role!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Role has been deleted!",
    )


class RoleManagerService(ImmediateNodeServiceWithReply):
    msg_handler_map: Dict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
        CreateRoleMessage: create_role_msg,
        UpdateRoleMessage: update_role_msg,
        GetRoleMessage: get_role_msg,
        GetRolesMessage: get_all_roles_msg,
        DeleteRoleMessage: del_role_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: INPUT_MESSAGES,
        verify_key: VerifyKey,
    ) -> OUTPUT_MESSAGES:
        return RoleManagerService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateRoleMessage,
            UpdateRoleMessage,
            GetRoleMessage,
            GetRolesMessage,
            DeleteRoleMessage,
        ]



# third party
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer

# relative
from . import Base


class UserGroup(Base):
    __tablename__ = "usergroup"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    user = Column(Integer, ForeignKey("syft_user.id"))
    group = Column(Integer, ForeignKey("group.id"))

    def __str__(self) -> str:
        return f"<UserGroup id: {self.id}, user: {self.user}, " f"group: {self.group}>"



"""Crypto Primitives."""

# stdlib
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import DefaultDict
from typing import Dict
from typing import List


class CryptoPrimitiveProvider:
    """A trusted third party should use this class to generate crypto primitives."""

    _func_providers: Dict[str, Callable] = {}
    _ops_list: DefaultDict[str, List] = defaultdict(list)

    def __init__(self) -> None:  # noqa
        raise ValueError("This class should not be initialized")

    @staticmethod
    def generate_primitives(
        op_str: str,
        parties: List[Any],
        g_kwargs: Dict[str, Any] = {},
        p_kwargs: Dict[str, Any] = {},
        nr_instances: int = 1,
        ring_size: int = 2 ** 32,
    ) -> List[Any]:
        """Generate "op_str" primitives.

        Args:
            op_str (str): Operator.
            parties (List[Any]): Parties to generate primitives for.
            g_kwargs: Generate kwargs passed to the registered function.
            p_kwargs: Populate kwargs passed to the registered populate function.

        Returns:
            List[Any]: List of primitives.

        Raises:
            ValueError: If op_str is not registered.

        """
        if op_str not in CryptoPrimitiveProvider._func_providers:
            raise ValueError(f"{op_str} not registered")

        nr_parties = len(parties)

        generator = CryptoPrimitiveProvider._func_providers[op_str]
        primitives = generator(
            **g_kwargs,
            nr_instances=nr_instances,
            nr_parties=nr_parties,
            ring_size=ring_size,
        )

        if p_kwargs is not None:
            """Do not transfer the primitives if there is not specified a
            values for populate kwargs."""
            CryptoPrimitiveProvider._transfer_primitives_to_parties(
                op_str=op_str,
                primitives=primitives,
                parties=parties,
                p_kwargs=p_kwargs,
                ring_size=ring_size,
            )

        # Since we do not have (YET!) the possiblity to return typed tuples from a remote
        # execute function we are using this.
        return primitives

    @staticmethod
    def _transfer_primitives_to_parties(
        op_str: str,
        primitives: List[Any],
        parties: List[Any],
        p_kwargs: Dict[str, Any],
        ring_size: int,
    ) -> None:
        if not isinstance(primitives, list):
            raise ValueError("Primitives should be a List")

        if len(primitives) != len(parties):
            raise ValueError(
                f"Primitives length {len(primitives)} != Parties length {len(parties)}"
            )

        for primitives_party, party in zip(primitives, parties):
            party.syft.core.tensor.smpc.share_tensor.populate_store(
                op_str, primitives_party, **p_kwargs, ring_size=ring_size
            )

    @staticmethod
    def get_state() -> str:
        """Get the state of a CryptoProvider.

        Returns:
            str: CryptoProvider
        """
        res = f"Providers: {list(CryptoPrimitiveProvider._func_providers.keys())}\n"
        return res



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/beaver_action.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.tensor import (
    share_tensor_pb2 as proto_dot_core_dot_tensor_dot_share__tensor__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n1proto/core/node/common/action/beaver_action.proto\x12\x1csyft.core.node.common.action\x1a$proto/core/tensor/share_tensor.proto\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\"\xe0\x01\n\x0c\x42\x65\x61verAction\x12*\n\x03\x65ps\x18\x01 \x01(\x0b\x32\x1d.syft.core.tensor.ShareTensor\x12%\n\x06\x65ps_id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID\x12,\n\x05\x64\x65lta\x18\x03 \x01(\x0b\x32\x1d.syft.core.tensor.ShareTensor\x12'\n\x08\x64\x65lta_id\x18\x04 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3"
)


_BEAVERACTION = DESCRIPTOR.message_types_by_name["BeaverAction"]
BeaverAction = _reflection.GeneratedProtocolMessageType(
    "BeaverAction",
    (_message.Message,),
    {
        "DESCRIPTOR": _BEAVERACTION,
        "__module__": "proto.core.node.common.action.beaver_action_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.BeaverAction)
    },
)
_sym_db.RegisterMessage(BeaverAction)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _BEAVERACTION._serialized_start = 190
    _BEAVERACTION._serialized_end = 414
# @@protoc_insertion_point(module_scope)



# stdlib
# stdlib
from typing import Any

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# relative
from ....proto.core.tensor.party_pb2 import Party as Party_PB
from ...common.serde.serializable import serializable


@serializable()
class Party:
    __slots__ = ("url", "port")
    # _DOCKER_HOST: str = "http://docker-host"

    def __init__(self, url: str, port: int) -> None:
        # TODO: This is not used -- it is hardcoded to docker
        # GM: Probably for real life scenario we would need to change this (when we serialize)
        # for a more real life scenario
        self.url = url

        self.port = port

    def _object2proto(self) -> Party_PB:
        return Party_PB(url=self.url, port=self.port)

    @staticmethod
    def _proto2object(proto: Party_PB) -> "Party":
        # TODO: If on the same machine use docker-host - if not real address
        # How to distinguish? (if 127.0.0.1 and localhost we consider using docker-host?)
        res = Party(url=proto.url, port=proto.port)
        return res

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Party_PB

    def __hash__(self) -> int:
        # TODO: Rasswanth, George, Trask this takes into consideration a hashing based on url and port
        # that we login only once
        # Is that sufficient?
        # res_str = f"{self.url}:{self.port}"
        return hash((self.url, self.port))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Party):
            return False

        return self.port == other.port and self.url == other.url



# relative
from ..abstract.node import AbstractNode
from ..common.node_manager.node_manager import NodeManager
from ..common.node_manager.node_route_manager import NodeRouteManager


class NodeServiceInterface(AbstractNode):
    node: NodeManager
    node_route: NodeRouteManager



# stdlib
from typing import Any
from typing import Dict
from typing import Tuple as TypeTuple

# third party
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

# relative
from . import Base
from .groups import Group
from .roles import Role
from .user import SyftUser
from .usergroup import UserGroup

datetime_cols = ["date", "created_at", "destroyed_at", "deployed_on", "updated_on"]


def model_to_json(model: Base) -> Dict[str, Any]:
    """Returns a JSON representation of an SQLAlchemy-backed object."""
    json = {}
    for col in model.__mapper__.attrs.keys():  # type: ignore
        if col != "hashed_password" and col != "salt":
            if col in datetime_cols:
                # Cast datetime object to string
                json[col] = str(getattr(model, col))
            else:
                json[col] = getattr(model, col)
    return json


def expand_user_object(_user: SyftUser, db: Engine) -> Dict[str, Any]:
    def get_group(user_group: UserGroup) -> Dict:
        query = db.session().query
        group = user_group.group
        group = query(Group).get(group)
        group = model_to_json(group)
        return group

    query = db.session().query
    user = model_to_json(_user)
    user["role"] = query(Role).get(user["role"])
    user["role"] = model_to_json(user["role"])
    user["groups"] = [
        get_group(user_group)
        for user_group in query(UserGroup).filter_by(user=user["id"]).all()
    ]
    return user


def seed_db(db: Session) -> None:
    new_role = Role(
        name="Data Scientist",
        can_make_data_requests=True,
        can_triage_data_requests=False,
        can_manage_privacy_budget=False,
        can_create_users=False,
        can_manage_users=False,
        can_edit_roles=False,
        can_manage_infrastructure=False,
        can_upload_data=False,
        can_upload_legal_document=False,
        can_edit_domain_settings=False,
    )
    db.add(new_role)

    new_role = Role(
        name="Compliance Officer",
        can_make_data_requests=False,
        can_triage_data_requests=True,
        can_manage_privacy_budget=False,
        can_create_users=False,
        can_manage_users=True,
        can_edit_roles=False,
        can_manage_infrastructure=False,
        can_upload_data=False,
        can_upload_legal_document=False,
        can_edit_domain_settings=False,
    )
    db.add(new_role)

    new_role = Role(
        name="Administrator",
        can_make_data_requests=True,
        can_triage_data_requests=True,
        can_manage_privacy_budget=True,
        can_create_users=True,
        can_manage_users=True,
        can_edit_roles=True,
        can_manage_infrastructure=True,
        can_upload_data=True,
        can_upload_legal_document=True,
        can_edit_domain_settings=True,
    )
    db.add(new_role)

    new_role = Role(
        name="Owner",
        can_make_data_requests=True,
        can_triage_data_requests=True,
        can_manage_privacy_budget=True,
        can_create_users=True,
        can_manage_users=True,
        can_edit_roles=True,
        can_manage_infrastructure=True,
        can_upload_data=True,
        can_upload_legal_document=True,
        can_edit_domain_settings=True,
    )
    db.add(new_role)
    db.commit()


def create_memory_db_engine() -> TypeTuple[Engine, sessionmaker]:
    db_engine = create_engine("sqlite://", echo=False)
    Base.metadata.create_all(db_engine)  # type: ignore
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=db_engine)
    seed_db(SessionLocal())
    return db_engine, SessionLocal



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/tensor/party.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1dproto/core/tensor/party.proto\x12\x10syft.core.tensor""\n\x05Party\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\rb\x06proto3'
)


_PARTY = DESCRIPTOR.message_types_by_name["Party"]
Party = _reflection.GeneratedProtocolMessageType(
    "Party",
    (_message.Message,),
    {
        "DESCRIPTOR": _PARTY,
        "__module__": "proto.core.tensor.party_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.tensor.Party)
    },
)
_sym_db.RegisterMessage(Party)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _PARTY._serialized_start = 51
    _PARTY._serialized_end = 85
# @@protoc_insertion_point(module_scope)



# # stdlib
# from typing import List
# from typing import Optional
#
# # third party
# from nacl.signing import VerifyKey
#
# # relative
# from ......logger import traceback_and_raise
# from ....abstract.node import AbstractNode
# from ....common.node_service.node_service import ImmediateNodeServiceWithoutReply
# from ..request_receiver.request_receiver_messages import RequestMessage
# from .get_all_requests_messages import GetAllRequestsMessage
# from .get_all_requests_messages import GetAllRequestsResponseMessage
#
#
# class GetAllRequestsService(ImmediateNodeServiceWithoutReply):
#     @staticmethod
#     def message_handler_types() -> List[type]:
#         return [GetAllRequestsMessage]
#
#     @staticmethod
#     def process(
#         node: AbstractNode,
#         msg: GetAllRequestsMessage,
#         verify_key: Optional[VerifyKey] = None,
#     ) -> GetAllRequestsResponseMessage:
#         try:
#             if verify_key is None:
#                 traceback_and_raise(
#                     ValueError(
#                         "Can't process Request service without a given " "verification key"
#                     )
#                 )
#
#             if verify_key == node.root_verify_key:
#                 return GetAllRequestsResponseMessage(
#                     requests=node.requests, address=msg.reply_to
#                 )
#
#             # only return requests which concern the user asking
#             valid_requests: List[RequestMessage] = list()
#             for request in node.requests:
#                 if request.requester_verify_key == verify_key:
#                     valid_requests.append(request)
#
#             return GetAllRequestsResponseMessage(
#                 requests=valid_requests, address=msg.reply_to
#             )
#         except Exception as e:
#             print('\n\nSOMETHING WENT WRONG!!!\n\n')
#             print(e)



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Union

# third party
import pandas as pd
import requests

# relative
from . import login
from .core.node.common.client import Client
from .logger import error
from .logger import warning

NETWORK_REGISTRY_URL = (
    "https://raw.githubusercontent.com/OpenMined/NetworkRegistry/main/networks.json"
)
NETWORK_REGISTRY_REPO = "https://github.com/OpenMined/NetworkRegistry"


class NetworkRegistry:
    def __init__(self) -> None:
        self.networks = []
        try:
            response = requests.get(NETWORK_REGISTRY_URL)
            network_json = response.json()
            self.networks = network_json["networks"]
        except Exception as e:
            warning(
                f"Failed to get Network Registry, go checkout: {NETWORK_REGISTRY_REPO}. {e}"
            )

    def _repr_html_(self) -> str:
        return pd.DataFrame(self.networks)._repr_html_()

    def create_client(self, network: Dict[str, Any]) -> Client:
        try:
            host_or_ip = network["host_or_ip"]
            port = int(network["port"])
            protocol = network["protocol"]
            return login(url=f"{protocol}://{host_or_ip}", port=port)
        except Exception as e:
            error(f"Failed to login with: {network}. {e}")
            raise e

    def __getitem__(self, key: Union[str, int]) -> Client:
        if isinstance(key, int):
            return self.create_client(network=self.networks[key])
        else:
            for network in self.networks:
                if network["name"] == key:
                    return self.create_client(network=network)
        raise KeyError(f"Invalid key: {key} for {self.networks}")



# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class Group(Base):

    __tablename__ = "group"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    name = Column(String(255))

    def __str__(self) -> str:
        return f"<Group id: {self.id}, name: {self.name}>"



"""A simple garbage collection heuritics."""
# stdlib
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

# third party
from typing_extensions import final

# relative
from ...common.uid import UID
from ...node.common.action.garbage_collect_batched_action import (
    GarbageCollectBatchedAction,
)
from ..pointer import Pointer
from .gc_strategy import GCStrategy

if TYPE_CHECKING:

    # relative
    from ...node.common.client import Client


@final
class GCBatched(GCStrategy):
    """The GCBatched Strategy."""

    __slots__ = [
        "obj_ids",
        "threshold",
        "client",
    ]

    client: Optional["Client"]
    obj_ids: List[UID]
    threshold: int

    def __init__(self, threshold: int = 10) -> None:
        """Construct the GCBatched Strategy.

        Args:
            threshold (int): the threshold after which a message
                would be sent to delete all the objects that were cached
        Return:
            None
        """
        self.obj_ids = []
        self.threshold = threshold
        self.client = None

    def reap(self, pointer: Pointer) -> None:
        """
        Check if we passed the threshold of objects that we should cache.
        If yes, then send a message to delete the objects.
        If no, cache the object id to be deleted when there are more ids
        collected.

        Args:
            pointer (Pointer): Pointer to the object that should get deleted

        Return:
            None
        """

        self.obj_ids.append(pointer.id_at_location)

        nr_objs_client = len(self.obj_ids)

        if nr_objs_client >= self.threshold:
            # Check the local threshold for the items that are kept for a client
            msg = GarbageCollectBatchedAction(
                ids_at_location=self.obj_ids, address=pointer.client.address
            )

            pointer.client.send_eventual_msg_without_reply(msg)
            self.obj_ids = []

        self.client = pointer.client

    def __del__(self) -> None:
        """Send a GarbageCollectBatchedAction to the client such that all the
        objects that are cached to be deleted would be deleted.
        """
        if self.client is None:
            return

        msg = GarbageCollectBatchedAction(
            ids_at_location=self.obj_ids, address=self.client.address
        )

        self.client.send_eventual_msg_without_reply(msg)



# relative
from .passthrough import AcceptableSimpleType  # type: ignore # noqa: 401
from .passthrough import PassthroughTensor  # type: ignore # noqa: 401
from .passthrough import SupportedChainType  # type: ignore  # noqa: 401



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/domain/service/get_all_requests_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.node.domain.service import (
    request_message_pb2 as proto_dot_core_dot_node_dot_domain_dot_service_dot_request__message__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n=proto/core/node/domain/service/get_all_requests_message.proto\x12\x1dsyft.core.node.domain.service\x1a\x34proto/core/node/domain/service/request_message.proto\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x8f\x01\n\x15GetAllRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\xaf\x01\n\x1dGetAllRequestsResponseMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12?\n\x08requests\x18\x03 \x03(\x0b\x32-.syft.core.node.domain.service.RequestMessageb\x06proto3'
)


_GETALLREQUESTSMESSAGE = DESCRIPTOR.message_types_by_name["GetAllRequestsMessage"]
_GETALLREQUESTSRESPONSEMESSAGE = DESCRIPTOR.message_types_by_name[
    "GetAllRequestsResponseMessage"
]
GetAllRequestsMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTSMESSAGE,
        "__module__": "proto.core.node.domain.service.get_all_requests_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.GetAllRequestsMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestsMessage)

GetAllRequestsResponseMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestsResponseMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTSRESPONSEMESSAGE,
        "__module__": "proto.core.node.domain.service.get_all_requests_message_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.GetAllRequestsResponseMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestsResponseMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _GETALLREQUESTSMESSAGE._serialized_start = 219
    _GETALLREQUESTSMESSAGE._serialized_end = 362
    _GETALLREQUESTSRESPONSEMESSAGE._serialized_start = 365
    _GETALLREQUESTSRESPONSEMESSAGE._serialized_end = 540
# @@protoc_insertion_point(module_scope)



# Custom Exception
# Purposefully raise a custom error to retry the task in celery worker.
# Whenever celery worker receives this Exception, it retries the action
# for a specified duration


class RetriableError(Exception):
    pass


class ObjectNotInStore(RetriableError):
    pass



# future
from __future__ import annotations

# stdlib
import json
import os
import re
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from nacl.signing import VerifyKey
import requests
from typing_extensions import final

# relative
from .....common.serde.serializable import serializable
from ....abstract.node import AbstractNode
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage


@serializable(recursive_serde=True)
@final
class VPNConnectMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNConnectReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNConnectMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNConnectMessage
    message_reply_type = VPNConnectReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            host_or_ip = str(self.kwargs["host_or_ip"])
            if not host_or_ip.startswith("http"):
                host_or_ip = f"http://{host_or_ip}/vpn"

            vpn_auth_key = str(self.kwargs["vpn_auth_key"])

            status, error = connect_with_key(
                tailscale_host="http://tailscale:4000",
                headscale_host=host_or_ip,
                vpn_auth_key=vpn_auth_key,
            )
            if status:
                return {"status": "ok"}
            else:
                print("connect with key failed", error)
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


def connect_with_key(
    tailscale_host: str, headscale_host: str, vpn_auth_key: str
) -> Tuple[bool, str]:
    data = {
        "args": [
            "-login-server",
            f"{headscale_host}",
            "--reset",
            "--force-reauth",
            "--authkey",
            f"{vpn_auth_key}",
        ],
        "timeout": 60,
        "STACK_API_KEY": os.environ.get("STACK_API_KEY", None),
    }
    command_url = f"{tailscale_host}/commands/up"

    resp = requests.post(command_url, json=data)
    report = get_result(json=resp.json())
    report_dict = json.loads(report)

    if int(report_dict["returncode"]) == 0:
        return (True, "")
    else:
        return (False, report_dict.get("report", ""))


@serializable(recursive_serde=True)
@final
class VPNJoinMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNJoinReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNJoinMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNJoinMessage
    message_reply_type = VPNJoinReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            host_or_ip = str(self.kwargs["host_or_ip"])
            if not host_or_ip.startswith("http"):
                host_or_ip = f"http://{host_or_ip}"

            # can't import Network due to circular imports
            if type(node).__name__ == "Network":
                # we are already in the network and could be on the blocking backend api
                msg = VPNRegisterMessageWithReply(kwargs={}).to(
                    address=node.address, reply_to=node.address
                )
                reply = node.recv_immediate_msg_with_reply(msg=msg).message
                res_json = {}
                try:
                    res_json["vpn_auth_key"] = str(
                        reply.payload.kwargs.get("vpn_auth_key")  # type: ignore
                    )
                except Exception:  # nosec
                    pass
            else:
                res = requests.post(f"{host_or_ip}/api/v1/vpn/register")
                res_json = res.json()

            if "vpn_auth_key" not in res_json:
                print("Registration failed", res)
                return {"status": "error"}

            status, error = connect_with_key(
                tailscale_host="http://tailscale:4000",
                headscale_host=f"{host_or_ip}/vpn",
                vpn_auth_key=res_json["vpn_auth_key"],
            )

            if status:
                return {"status": "ok"}
            else:
                print("connect with key failed", error)
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


@serializable(recursive_serde=True)
@final
class VPNRegisterMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNRegisterReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNRegisterMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNRegisterMessage
    message_reply_type = VPNRegisterReplyMessage

    def run(self, node: AbstractNode, verify_key: Optional[VerifyKey] = None) -> Any:
        try:
            status, vpn_auth_key = generate_key(headscale_host="http://headscale:4000")

            if status:
                return {"status": "ok", "vpn_auth_key": vpn_auth_key}
            else:
                print("register and create vpn_auth_key failed")
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


# the response contains a dict with a key called "report" which contains unescaped json
def extract_nested_json(nested_json: str) -> Union[Dict, List]:
    matcher = r'"report":"(.+)\\n"'
    parts = re.findall(matcher, nested_json)
    report_json = parts[0].replace("\\n", "").replace("\\t", "").replace('\\"', '"')
    return json.loads(report_json)


def generate_key(headscale_host: str) -> Tuple[bool, str]:
    data = {
        "timeout": 5,
        "STACK_API_KEY": os.environ.get("STACK_API_KEY", None),
    }

    command_url = f"{headscale_host}/commands/generate_key"
    try:
        resp = requests.post(command_url, json=data)
        report = get_result(json=resp.json())
        result_dict = dict(extract_nested_json(report))
        result = result_dict["Key"]

        # check if we got a key
        if len(result) == 48:
            return (True, result)
        else:
            return (False, "")

    except Exception as e:
        print("failed to make request", e)
        return (False, str(e))


@serializable(recursive_serde=True)
@final
class VPNStatusMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNStatusReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNStatusMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNStatusMessage
    message_reply_type = VPNStatusReplyMessage

    def run(self, node: AbstractNode, verify_key: Optional[VerifyKey] = None) -> Any:
        try:
            up, host, peers = get_status(tailscale_host="http://tailscale:4000")
            return {"status": "ok", "connected": up, "host": host, "peers": peers}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


def clean_status_output(
    input: str,
) -> Tuple[bool, Dict[str, str], List[Dict[str, str]]]:
    # example input
    """
    # Health check:
    #     - dns: rename /etc/resolv.conf /etc/resolv.pre-tailscale-backup.conf: device or resource busy

    100.64.0.1      test_domain_1        omnet        linux   -
    100.64.0.2      test_network_1       omnet        linux   active; relay "syd", tx 1188 rx 1040
    """
    up = False
    peers: List[Dict[str, str]] = []
    host: Dict[str, str] = {}
    if "Tailscale is stopped." in input:
        return up, host, peers
    elif "unexpected state: NoState" in input:
        return up, host, peers

    count = 0
    for line in str(input).split("\n"):
        matches = re.match(r"^\d.+", line)
        if matches is not None:
            try:
                stat_parts = re.split(r"(\s+)", matches.string)
                entry = {}
                entry["ip"] = stat_parts[0]
                entry["hostname"] = stat_parts[2]
                entry["network"] = stat_parts[4]
                entry["os"] = stat_parts[6]
                connection_info_parts = matches.string.split(entry["os"])
                entry["connection_info"] = "n/a"
                connection_info = ""
                if len(connection_info_parts) > 1:
                    connection_info = connection_info_parts[1].strip()
                    entry["connection_info"] = connection_info

                entry["connection_status"] = "n/a"
                if "active" in connection_info:
                    entry["connection_status"] = "active"

                if "idle" in connection_info:
                    entry["connection_status"] = "idle"

                entry["connection_type"] = "n/a"
                if "relay" in connection_info:
                    entry["connection_type"] = "relay"

                if "direct" in connection_info:
                    entry["connection_type"] = "direct"

                if count == 0:
                    host = entry
                    count += 1
                    up = True
                else:
                    peers.append(entry)

            except Exception as e:
                print("Error parsing tailscale status output", e)
                pass

    return up, host, peers


def get_status(
    tailscale_host: str,
) -> Tuple[bool, Dict[str, str], List[Dict[str, str]]]:
    data = {
        "timeout": 5,
        "STACK_API_KEY": os.environ.get("STACK_API_KEY", None),
    }
    command_url = f"{tailscale_host}/commands/status"
    host: Dict[str, str] = {}
    peers: List[Dict[str, str]] = []
    connected = False
    try:
        resp = requests.post(command_url, json=data)
        report = get_result(json=resp.json())
        cmd_output = json.loads(report)["report"]
        connected, host, peers = clean_status_output(input=cmd_output)
    except Exception as e:
        print("failed to make request", e)
    return connected, host, peers


def get_result(json: Dict) -> str:
    result_url = json.get("result_url", "")
    result_url += f'&STACK_API_KEY={os.environ.get("STACK_API_KEY", None)}'
    tries = 0
    limit = 5
    try:
        while True:
            print("Polling API Result", tries)
            result = requests.get(result_url)
            if '"status":"running"' in result.text:
                time.sleep(1)
                tries += 1
                if tries > limit:
                    break
                continue
            else:
                return str(result.text)
    except Exception as e:
        print("Failed to get result", json, e)
    return "{}"



# stdlib
import os
from pathlib import Path
from secrets import randbelow
from typing import Any
from typing import List
from typing import Optional
from typing import Union

# third party
from forbiddenfruit import curse
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft absolute
import syft

# relative
from .logger import critical
from .logger import debug
from .logger import error
from .logger import traceback_and_raise


def validate_type(_object: object, _type: type, optional: bool = False) -> Any:
    if isinstance(_object, _type) or (optional and (_object is None)):
        return _object

    traceback_and_raise(
        f"Object {_object} should've been of type {_type}, not {_object}."
    )


def validate_field(_object: object, _field: str) -> Any:
    object = getattr(_object, _field, None)

    if object is not None:
        return object

    traceback_and_raise(f"Object {_object} has no {_field} field set.")


def get_subclasses(obj_type: type) -> List[type]:
    """Recursively generate the list of all classes within the sub-tree of an object

    As a paradigm in Syft, we often allow for something to be known about by another
    part of the codebase merely because it has subclassed a particular object. While
    this can be a big "magicish" it also can simplify future extensions and reduce
    the likelihood of small mistakes (if done right).

    This is a utility function which allows us to look for sub-classes and the sub-classes
    of those sub-classes etc. returning a full list of descendants of a class in the inheritance
    hierarchy.

    Args:
        obj_type: the type we want to look for sub-classes of

    Returns:
        the list of subclasses of obj_type

    """

    classes = list()
    for sc in obj_type.__subclasses__():
        classes.append(sc)
        classes += get_subclasses(obj_type=sc)
    return classes


def index_modules(a_dict: object, keys: List[str]) -> object:
    """Recursively find a syft module from its path

    This is the recursive inner function of index_syft_by_module_name.
    See that method for a full description.

    Args:
        a_dict: a module we're traversing
        keys: the list of string attributes we're using to traverse the module

    Returns:
        a reference to the final object

    """

    if len(keys) == 0:
        return a_dict
    return index_modules(a_dict=a_dict.__dict__[keys[0]], keys=keys[1:])


def index_syft_by_module_name(fully_qualified_name: str) -> object:
    """Look up a Syft class/module/function from full path and name

    Sometimes we want to use the fully qualified name (such as one
    generated from the 'get_fully_qualified_name' method below) to
    fetch an actual reference. This is most commonly used in deserialization
    so that we can have generic protobuf objects which just have a string
    representation of the specific object it is meant to deserialize to.

    Args:
        fully_qualified_name: the name in str of a module, class, or function

    Returns:
        a reference to the actual object at that string path

    """

    # @Tudor this needs fixing during the serde refactor
    # we should probably just support the native type names as lookups for serde
    if fully_qualified_name == "builtins.NoneType":
        fully_qualified_name = "syft.lib.python._SyNone"
    attr_list = fully_qualified_name.split(".")

    # we deal with VerifyAll differently, because we don't it be imported and used by users
    if attr_list[-1] == "VerifyAll":
        return type(syft.core.common.group.VERIFYALL)

    if attr_list[0] != "syft":
        raise ReferenceError(f"Reference don't match: {attr_list[0]}")

    if (
        attr_list[1] != "core"
        and attr_list[1] != "lib"
        and attr_list[1] != "grid"
        and attr_list[1] != "wrappers"
        and attr_list[1] != "proxy"
    ):
        raise ReferenceError(f"Reference don't match: {attr_list[1]}")

    return index_modules(a_dict=globals()["syft"], keys=attr_list[1:])


def get_fully_qualified_name(obj: object) -> str:
    """Return the full path and name of a class

    Sometimes we want to return the entire path and name encoded
    using periods. For example syft.core.common.message.SyftMessage
    is the current fully qualified path and name for the SyftMessage
    object.

    Args:
        obj: the object we want to get the name of

    Returns:
        the full path and name of the object

    """

    fqn = obj.__class__.__module__
    try:
        fqn += "." + obj.__class__.__name__
    except Exception as e:
        error(f"Failed to get FQN: {e}")
    return fqn


def aggressive_set_attr(obj: object, name: str, attr: object) -> None:
    """Different objects prefer different types of monkeypatching - try them all

    Args:
        obj: object whose attribute has to be set
        name: attribute name
        attr: value given to the attribute

    """
    try:
        setattr(obj, name, attr)
    except Exception:
        curse(obj, name, attr)


def obj2pointer_type(obj: object) -> type:
    fqn = None
    try:
        fqn = get_fully_qualified_name(obj=obj)
    except Exception as e:
        # sometimes the object doesn't have a __module__ so you need to use the type
        # like: collections.OrderedDict
        debug(f"Unable to get get_fully_qualified_name of {type(obj)} trying type. {e}")
        fqn = get_fully_qualified_name(obj=type(obj))

    # TODO: fix for other types
    if obj is None:
        fqn = "syft.lib.python._SyNone"

    try:
        ref = syft.lib_ast.query(fqn, obj_type=type(obj))
    except Exception as e:
        log = f"Cannot find {type(obj)} {fqn} in lib_ast. {e}"
        critical(log)
        raise Exception(log)

    return ref.pointer_type  # type: ignore


def key_emoji(key: object) -> str:
    try:
        if isinstance(key, (bytes, SigningKey, VerifyKey)):
            hex_chars = bytes(key).hex()[-8:]
            return char_emoji(hex_chars=hex_chars)
    except Exception as e:
        error(f"Fail to get key emoji: {e}")
        pass
    return "ALL"


def char_emoji(hex_chars: str) -> str:
    base = ord("\U0001F642")
    hex_base = ord("0")
    code = 0
    for char in hex_chars:
        offset = ord(char)
        code += offset - hex_base
    return chr(base + code)


left_name = [
    "admiring",
    "adoring",
    "affectionate",
    "agitated",
    "amazing",
    "angry",
    "awesome",
    "beautiful",
    "blissful",
    "bold",
    "boring",
    "brave",
    "busy",
    "charming",
    "clever",
    "cool",
    "compassionate",
    "competent",
    "condescending",
    "confident",
    "cranky",
    "crazy",
    "dazzling",
    "determined",
    "distracted",
    "dreamy",
    "eager",
    "ecstatic",
    "elastic",
    "elated",
    "elegant",
    "eloquent",
    "epic",
    "exciting",
    "fervent",
    "festive",
    "flamboyant",
    "focused",
    "friendly",
    "frosty",
    "funny",
    "gallant",
    "gifted",
    "goofy",
    "gracious",
    "great",
    "happy",
    "hardcore",
    "heuristic",
    "hopeful",
    "hungry",
    "infallible",
    "inspiring",
    "interesting",
    "intelligent",
    "jolly",
    "jovial",
    "keen",
    "kind",
    "laughing",
    "loving",
    "lucid",
    "magical",
    "mystifying",
    "modest",
    "musing",
    "naughty",
    "nervous",
    "nice",
    "nifty",
    "nostalgic",
    "objective",
    "optimistic",
    "peaceful",
    "pedantic",
    "pensive",
    "practical",
    "priceless",
    "quirky",
    "quizzical",
    "recursing",
    "relaxed",
    "reverent",
    "romantic",
    "sad",
    "serene",
    "sharp",
    "silly",
    "sleepy",
    "stoic",
    "strange",
    "stupefied",
    "suspicious",
    "sweet",
    "tender",
    "thirsty",
    "trusting",
    "unruffled",
    "upbeat",
    "vibrant",
    "vigilant",
    "vigorous",
    "wizardly",
    "wonderful",
    "xenodochial",
    "youthful",
    "zealous",
    "zen",
]

right_name = [
    "altman",
    "bach",
    "bengio",
    "bostrom",
    "botvinick",
    "brockman",
    "chintala",
    "chollet",
    "chomsky",
    "dean",
    "fridman",
    "gardner",
    "goertzel",
    "goodfellow",
    "hassabis",
    "he",
    "hinton",
    "hochreiter",
    "hotz",
    "howard",
    "hutter",
    "kaliouby",
    "karp",
    "karpathy",
    "kearns",
    "kellis",
    "knuth",
    "koller",
    "krizhevsky",
    "larochelle",
    "lecun",
    "li",
    "lim",
    "malik",
    "mironov",
    "ng",
    "norvig",
    "olah",
    "pearl",
    "pesenti",
    "russell",
    "salakhutdinov",
    "schmidhuber",
    "silver",
    "smola",
    "song",
    "sophia",
    "sutskever",
    "thomas",
    "thrun",
    "trask",
    "vapnik",
    "vaswani",
    "vinyals",
    "winston",
    "wolf",
    "wolfram",
]


def random_name() -> str:
    left_i = randbelow(len(left_name) - 1)
    right_i = randbelow(len(right_name) - 1)
    return f"{left_name[left_i].capitalize()} {right_name[right_i].capitalize()}"


def inherit_tags(
    attr_path_and_name: str,
    result: object,
    self_obj: Optional[object],
    args: Union[tuple, list],
    kwargs: dict,
) -> None:
    tags = []
    if self_obj is not None and hasattr(self_obj, "tags"):
        tags.extend(list(self_obj.tags))  # type: ignore

    for arg in args:
        if hasattr(arg, "tags"):
            tags.extend([tag for tag in arg.tags if tag not in tags])

    for arg in kwargs.values():
        if hasattr(arg, "tags"):
            tags.extend([tag for tag in arg.tags if tag not in tags])

    # only generate new tags if the result actually inherit some tags.
    if tags:
        tags.append(attr_path_and_name.split(".")[-1])
        result.tags = tags  # type: ignore


def get_root_data_path() -> Path:
    # get the PySyft / data directory to share datasets between notebooks
    # on Linux and MacOS the directory is: ~/.syft/data"
    # on Windows the directory is: C:/Users/$USER/.syft/data

    data_dir = Path.home() / ".syft" / "data"

    os.makedirs(data_dir, exist_ok=True)
    return data_dir



# future
from __future__ import annotations

# stdlib
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey

# relative
from ...core.common.uid import UID
from ...core.io.address import Address
from ...core.io.location import SpecificLocation
from ...core.io.route import SoloRoute
from ...core.node.common import AbstractNodeClient
from ...core.node.domain.client import DomainClient
from ...logger import error
from .grid_connection import GridHTTPConnection


class ProxyClient(DomainClient):
    def post_init(self) -> None:
        # use post_init so we don't need to re-implement an empty regular init
        super().post_init()
        self.logged_in = False

    @staticmethod
    def create(
        proxy_node_client: AbstractNodeClient,
        remote_domain: Union[str, UID, Address],
        domain_name: str,
    ) -> ProxyClient:
        domain_address = remote_domain
        try:
            if isinstance(domain_address, str):
                domain_address = UID.from_string(value=domain_address)
        except Exception as e:
            error(f"Failed to convert remote_domain str to UID. {e}")

        try:
            if isinstance(domain_address, UID):
                spec_location = SpecificLocation(domain_address)
                domain_address = Address(name=domain_name, domain=spec_location)
        except Exception as e:
            error(f"Failed to convert remote_domain UID to Address. {e}")

        if not isinstance(domain_address, Address):
            raise Exception(
                f"Failed to convert remote_domain {domain_address} to Address"
            )

        # the DomainRequestAPI is only on Network Clients so here we can re-use
        # the existing client route which will be pointing to the current Network
        base_url = proxy_node_client.routes[0].connection.base_url  # type: ignore
        conn = GridHTTPConnection(url=base_url)

        # this will initially be a guest connection so we should generate a new
        # unique key
        _user_key = SigningKey.generate()

        # Create a new Solo Route using the selected connection type
        route = SoloRoute(destination=spec_location, connection=conn)

        # here we mix all of the above together into a ProxyClient which is like a
        # normal client, except the route and the address are for different nodes
        proxy_client = ProxyClient(
            name=domain_name,
            routes=[route],
            signing_key=_user_key,
            domain=domain_address.domain,  # type: ignore
        )

        return proxy_client

    def login(self, email: str, password: str) -> ProxyClient:
        result = self.users.login(email=email, password=password)
        try:
            if result["status"] == "ok":
                _user_key = SigningKey(
                    result["data"]["key"].encode(), encoder=HexEncoder
                )
                self.signing_key = _user_key
                self.verify_key = self.signing_key.verify_key
                print(f"Logged in to {self.name} as {email}")
                self.logged_in = True
            else:
                print(f"Failed to login as {email}")
        except Exception as e:
            print(f"Failed to decode private key. {e}")
        return self

    def logout(self) -> ProxyClient:
        print(f"Logged out of account. You are now a Guest on {self.name}")
        self.signing_key = SigningKey.generate()
        self.verify_key = self.signing_key.verify_key
        self.logged_in = False
        return self



# stdlib
from typing import Set as TypeSet
from uuid import UUID

# relative
from .....lib.python import Set
from ....common import UID
from ....common.message import SignedImmediateSyftMessageWithoutReply
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from .exceptions import RetriableError

# Setting a deterministic location to access the unfinished tasks.
# TODO: Should modify it to to be able to parallel processin on the unfinished tasks
id = UID(UUID("a8ac0c37382584a1082c710b0b38f6a3"))


def get_set(node: AbstractNode) -> TypeSet:
    obj = node.store.get_object(key=id)
    if obj is None:
        return Set(set())
    elif isinstance(obj.data, Set):
        return obj.data
    else:
        raise ValueError(
            f"Unfinished task Object should be {obj} should be a Set or None"
        )


def update_set(
    task_set: TypeSet[SignedImmediateSyftMessageWithoutReply], node: AbstractNode
) -> None:
    obj = StorableObject(data=task_set, id=id)
    node.store[id] = obj


def register_unfinished_task(
    message: SignedImmediateSyftMessageWithoutReply, node: AbstractNode
) -> None:
    UNFINISHED_TASKS: TypeSet[SignedImmediateSyftMessageWithoutReply] = get_set(node)
    UNFINISHED_TASKS.add(message)
    update_set(UNFINISHED_TASKS, node)


def proceed_unfinished_tasks(node: AbstractNode) -> None:
    UNFINISHED_TASKS: TypeSet[SignedImmediateSyftMessageWithoutReply] = get_set(node)
    for unfinished_task in UNFINISHED_TASKS.copy():
        try:
            node.recv_immediate_msg_without_reply(unfinished_task)
            UNFINISHED_TASKS.remove(unfinished_task)
        except Exception as e:
            if not isinstance(e, RetriableError):
                raise e
    update_set(UNFINISHED_TASKS, node)



# stdlib
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from bcrypt import checkpw
from bcrypt import gensalt
from bcrypt import hashpw
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pydantic import BaseModel
from pydantic import EmailStr
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Query
from sqlalchemy.orm import sessionmaker

# relative
from ..exceptions import InvalidCredentialsError
from ..exceptions import UserNotFoundError
from ..node_table.pdf import PDFObject
from ..node_table.roles import Role
from ..node_table.user import SyftUser
from ..node_table.user import UserApplication
from .database_manager import DatabaseManager
from .role_manager import RoleManager


# Shared properties
class UserBase(BaseModel):
    email: Optional[EmailStr] = None
    is_active: Optional[bool] = True
    is_superuser: bool = False
    full_name: Optional[str] = None


# Properties to receive via API on creation
class UserCreate(UserBase):
    email: EmailStr
    password: str


# Properties to receive via API on update
class UserUpdate(UserBase):
    password: Optional[str] = None


class UserInDBBase(UserBase):
    id: Optional[int] = None

    class Config:
        orm_mode = True


# Additional properties to return via API
class User(UserInDBBase):
    pass


# Additional properties stored in DB
class UserInDB(UserInDBBase):
    hashed_password: str


class UserManager(DatabaseManager):

    schema = SyftUser

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=UserManager.schema, db=database)
        self.roles = RoleManager(database)

    @property
    def common_users(self) -> list:
        common_users: List[SyftUser] = []
        for role in self.roles.common_roles:
            common_users = common_users + list(super().query(role=role.id))

        return common_users

    @property
    def org_users(self) -> list:
        org_users: List[SyftUser] = []
        for role in self.roles.org_roles:
            org_users = org_users + list(super().query(role=role.id))
        return org_users

    def create_user_application(
        self,
        name: str,
        email: str,
        password: str,
        daa_pdf: Optional[bytes],
        institution: Optional[str] = "",
        website: Optional[str] = "",
        budget: Optional[float] = 0.0,
    ) -> int:
        salt, hashed = self.__salt_and_hash_password(password, 12)
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        _pdf_obj = PDFObject(binary=daa_pdf)
        session_local.add(_pdf_obj)
        session_local.commit()
        session_local.flush()
        session_local.refresh(_pdf_obj)

        _obj = UserApplication(
            name=name,
            email=email,
            salt=salt,
            hashed_password=hashed,
            daa_pdf=_pdf_obj.id,
            institution=institution,
            website=website,
            budget=budget,
        )
        session_local.add(_pdf_obj)
        session_local.add(_obj)
        session_local.commit()
        _obj_id = _obj.id
        session_local.close()
        return _obj_id

    def get_all_applicant(self) -> List[UserApplication]:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        result = list(session_local.query(UserApplication).all())
        session_local.close()
        return result

    def process_user_application(
        self, candidate_id: int, status: str, verify_key: VerifyKey
    ) -> None:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        candidate = (
            session_local.query(UserApplication).filter_by(id=candidate_id).first()
        )
        session_local.close()

        if status == "accepted":
            # Generate a new signing key
            _private_key = SigningKey.generate()

            encoded_pk = _private_key.encode(encoder=HexEncoder).decode("utf-8")
            encoded_vk = _private_key.verify_key.encode(encoder=HexEncoder).decode(
                "utf-8"
            )
            added_by = self.get_user(verify_key).name  # type: ignore
            self.register(
                name=candidate.name,
                email=candidate.email,
                role=self.roles.ds_role.id,
                budget=candidate.budget,
                private_key=encoded_pk,
                verify_key=encoded_vk,
                hashed_password=candidate.hashed_password,
                salt=candidate.salt,
                daa_pdf=candidate.daa_pdf,
                added_by=added_by,
                institution=candidate.institution,
                website=candidate.website,
                created_at=datetime.now(),
            )
        else:
            status = "rejected"

        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        candidate = (
            session_local.query(UserApplication).filter_by(id=candidate_id).first()
        )
        candidate.status = status
        session_local.flush()
        session_local.commit()
        session_local.close()

    def signup(
        self,
        name: str,
        email: str,
        password: str,
        budget: float,
        role: int,
        private_key: str,
        verify_key: str,
    ) -> SyftUser:
        salt, hashed = self.__salt_and_hash_password(password, 12)
        return self.register(
            name=name,
            email=email,
            role=role,
            budget=budget,
            private_key=private_key,
            verify_key=verify_key,
            hashed_password=hashed,
            salt=salt,
            created_at=datetime.now(),
        )

    def query(self, **kwargs: Any) -> Query:
        results = super().query(**kwargs)
        return results

    def first(self, **kwargs: Any) -> SyftUser:
        result = super().first(**kwargs)
        if not result:
            raise UserNotFoundError
        return result

    def login(self, email: str, password: str) -> SyftUser:
        return self.__login_validation(email, password)

    def set(  # nosec
        self,
        user_id: str,
        email: str = "",
        password: str = "",
        role: int = 0,
        name: str = "",
        website: str = "",
        institution: str = "",
        budget: float = 0.0,
    ) -> None:
        if not self.contain(id=user_id):
            raise UserNotFoundError

        key: str
        value: Union[str, int, float]

        if email:
            key = "email"
            value = email
        elif password:
            salt, hashed = self.__salt_and_hash_password(password, 12)
            self.modify({"id": user_id}, {"salt": salt, "hashed_password": hashed})
            return
        elif role != 0:
            key = "role"
            value = role
        elif name:
            key = "name"
            value = name
        elif budget:
            key = "budget"
            value = budget
        elif website:
            key = "website"
            value = website
        elif budget:
            key = "budget"
            value = budget
        elif institution:
            key = "institution"
            value = institution
        else:
            raise Exception

        self.modify({"id": user_id}, {key: value})

    def can_create_users(self, verify_key: VerifyKey) -> bool:
        try:
            return self.role(verify_key=verify_key).can_create_users
        except UserNotFoundError:
            return False

    def can_upload_data(self, verify_key: VerifyKey) -> bool:
        try:
            return self.role(verify_key=verify_key).can_upload_data
        except UserNotFoundError:
            return False

    def can_triage_requests(self, verify_key: VerifyKey) -> bool:
        try:
            return self.role(verify_key=verify_key).can_triage_data_requests
        except UserNotFoundError:
            return False

    def can_manage_infrastructure(self, verify_key: VerifyKey) -> bool:
        try:
            return self.role(verify_key=verify_key).can_manage_infrastructure
        except UserNotFoundError:
            return False

    def can_edit_roles(self, verify_key: VerifyKey) -> bool:
        try:
            return self.role(verify_key=verify_key).can_edit_roles
        except UserNotFoundError:
            return False

    def role(self, verify_key: VerifyKey) -> Role:
        user = self.get_user(verify_key)
        if not user:
            raise UserNotFoundError
        return self.roles.first(id=user.role)

    def get_user(self, verify_key: VerifyKey) -> Optional[SyftUser]:
        return self.first(
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
        )

    def __login_validation(self, email: str, password: str) -> SyftUser:
        try:
            user = self.first(email=email)
            if not user:
                raise UserNotFoundError

            hashed = user.hashed_password.encode("UTF-8")
            salt = user.salt.encode("UTF-8")
            bytes_pass = password.encode("UTF-8")

            if checkpw(bytes_pass, salt + hashed):
                return user
            else:
                raise InvalidCredentialsError
        except UserNotFoundError:
            raise InvalidCredentialsError

    def __salt_and_hash_password(self, password: str, rounds: int) -> Tuple[str, str]:
        bytes_pass = password.encode("UTF-8")
        salt = gensalt(rounds=rounds)
        salt_len = len(salt)
        hashed = hashpw(bytes_pass, salt)
        hashed = hashed[salt_len:]
        hashed = hashed.decode("UTF-8")
        salt = salt.decode("UTF-8")
        return salt, hashed



# stdlib
from typing import Iterable
from typing import KeysView
from typing import List
from typing import Optional
from typing import cast

# third party
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
from torch import Tensor

# syft absolute
import syft

# relative
from ....common.uid import UID
from ....node.common.node_table.bin_obj_dataset import BinObjDataset
from ....store import ObjectStore
from ....store.storeable_object import StorableObject
from ..node_table.bin_obj import BinObject
from ..node_table.bin_obj_metadata import ObjectMetadata

ENCODING = "UTF-8"


def create_storable(
    _id: UID, data: Tensor, description: str, tags: Optional[List[str]] = None
) -> StorableObject:
    obj = StorableObject(id=_id, data=data, description=description, tags=tags)

    return obj


class BinObjectManager(ObjectStore):
    def __init__(self, db: Session) -> None:
        self.db = db

    def get_object(self, key: UID) -> Optional[StorableObject]:
        try:
            return self.__getitem__(key)
        except KeyError as e:  # noqa: F841
            return None

    def get_objects_of_type(self, obj_type: type) -> Iterable[StorableObject]:
        return [obj for obj in self.values() if isinstance(obj.data, obj_type)]

    def __sizeof__(self) -> int:
        return self.values().__sizeof__()

    def __str__(self) -> str:
        return str(self.values())

    def __len__(self) -> int:
        local_session = sessionmaker(bind=self.db)()
        result = local_session.query(ObjectMetadata).count()
        local_session.close()
        return result

    def keys(self) -> KeysView[UID]:
        local_session = sessionmaker(bind=self.db)()
        keys = local_session.query(BinObject.id).all()
        keys = [UID.from_string(k[0]) for k in keys]
        local_session.close()
        return keys

    def values(self) -> List[StorableObject]:

        obj_keys = self.keys()
        values = []
        for key in obj_keys:
            try:
                values.append(self.__getitem__(key))
            except Exception as e:  # noqa: F841
                print("Unable to get item for key", key)  # TODO: TechDebt add logging
                print(e)
        return values

    def __contains__(self, key: UID) -> bool:
        local_session = sessionmaker(bind=self.db)()
        result = (
            local_session.query(BinObject).filter_by(id=str(key.value)).first()
            is not None
        )
        local_session.close()
        return result

    def __getitem__(self, key: UID) -> StorableObject:
        local_session = sessionmaker(bind=self.db)()
        bin_obj = local_session.query(BinObject).filter_by(id=str(key.value)).first()
        obj_metadata = (
            local_session.query(ObjectMetadata).filter_by(obj=str(key.value)).first()
        )

        if not bin_obj or not obj_metadata:
            raise KeyError(f"Object not found! for UID: {key}")

        obj = StorableObject(
            id=UID.from_string(bin_obj.id),
            data=bin_obj.obj,
            description=obj_metadata.description,
            tags=obj_metadata.tags,
            read_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.read_permissions), from_bytes=True
                )
            ),
            search_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.search_permissions), from_bytes=True
                )
            ),
            # name=obj_metadata.name,
        )
        local_session.close()
        return obj

    def is_dataset(self, key: UID) -> bool:
        local_session = sessionmaker(bind=self.db)()
        is_dataset_obj = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).exists()
        )
        is_dataset_obj = local_session.query(is_dataset_obj).scalar()
        local_session.close()
        return is_dataset_obj

    def _get_obj_dataset_relation(self, key: UID) -> Optional[BinObjDataset]:
        local_session = sessionmaker(bind=self.db)()
        obj_dataset_relation = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).first()
        )
        local_session.close()
        return obj_dataset_relation

    def __setitem__(self, key: UID, value: StorableObject) -> None:
        bin_obj = BinObject(id=str(key.value), obj=value.data)
        # metadata_dict = storable_to_dict(value)
        metadata_obj = ObjectMetadata(
            obj=bin_obj.id,
            tags=value.tags,
            description=value.description,
            read_permissions=cast(
                bytes,
                syft.serialize(
                    syft.lib.python.Dict(value.read_permissions), to_bytes=True
                ),
            ).hex(),
            search_permissions=cast(
                bytes,
                syft.serialize(
                    syft.lib.python.Dict(value.search_permissions), to_bytes=True
                ),
            ).hex(),
            # name=metadata_dict["name"],
        )

        obj_dataset_relation = self._get_obj_dataset_relation(key)
        if obj_dataset_relation:
            # Create a object dataset relationship for the new object
            obj_dataset_relation = BinObjDataset(
                id=obj_dataset_relation.id,
                name=obj_dataset_relation.name,
                obj=bin_obj.id,
                dataset=obj_dataset_relation.dataset,
                dtype=obj_dataset_relation.dtype,
                shape=obj_dataset_relation.shape,
            )

        if self.__contains__(key):
            self.delete(key)

        local_session = sessionmaker(bind=self.db)()
        local_session.add(bin_obj)
        local_session.add(metadata_obj)
        local_session.add(obj_dataset_relation) if obj_dataset_relation else None
        local_session.commit()
        local_session.close()

    def delete(self, key: UID) -> None:
        try:
            local_session = sessionmaker(bind=self.db)()

            object_to_delete = (
                local_session.query(BinObject).filter_by(id=str(key.value)).first()
            )
            metadata_to_delete = (
                local_session.query(ObjectMetadata)
                .filter_by(obj=str(key.value))
                .first()
            )
            local_session.delete(metadata_to_delete)
            local_session.delete(object_to_delete)
            local_session.commit()
            local_session.close()
        except Exception as e:
            print(f"{type(self)} Exception in __delitem__ error {key}. {e}")

    def clear(self) -> None:
        local_session = sessionmaker(bind=self.db)()
        local_session.query(BinObject).delete()
        local_session.query(ObjectMetadata).delete()
        local_session.commit()
        local_session.close()

    def __repr__(self) -> str:
        return str(self.values())



class ExperimentalFlags:
    def __init__(self) -> None:
        self._APACHE_ARROW_TENSOR_SERDE = True

    @property
    def APACHE_ARROW_TENSOR_SERDE(self) -> bool:
        return self._APACHE_ARROW_TENSOR_SERDE

    @APACHE_ARROW_TENSOR_SERDE.getter
    def APACHE_ARROW_TENSOR_SERDE(self) -> bool:
        return self._APACHE_ARROW_TENSOR_SERDE

    @APACHE_ARROW_TENSOR_SERDE.setter
    def APACHE_ARROW_TENSOR_SERDE(self, value: bool) -> None:
        self._APACHE_ARROW_TENSOR_SERDE = value


flags = ExperimentalFlags()
